//
//   cfg.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: cfg.c,v 1.17 1998/12/04 09:04:35 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: cfg.c,v 1.17 1998/12/04 09:04:35 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>

#include "cfg.h"
#include "utils.h"
#include "getlopt.h"

#ifndef GLOBAL_ETC_DIR
#define GLOBAL_ETC_DIR "/usr/local/etc"
#endif

static int
cfg_perform (const topt *opt, char *val)
{
	if (!(opt->flags & 1)) {	// doesn't take argument
		if (opt->var)
			if (opt->flags & 2) /* var is *char */
				*((char *) opt->var) = (char) opt->value;
			else
				*((int *) opt->var) = opt->value;
	}
	else {	// requires argument
		if (opt->var)
			if (opt->flags & 2) { /* var is *char */
				if (!(*((char **) opt->var) = strdup(val)))
					out_of_memory();
			}
			else {
				char *eptr;

				*((int *) opt->var) = strtol(val, &eptr, 0);
				if (*val && !*eptr) {
					fprintf (stderr, "%s: \"%s\": "
					    "illegal integer value\n",
					    me, val);
					return -1;
				}
			}
	}
	if (opt->func)
		return opt->func(val);
	return 0;
}

static int
cfg_set (const topt *opts, char *key, char *val)
{
	const topt *opt;

	for (opt = opts; opt->sname || opt->lname; opt++)
		if (opt->lname && !strcmp(opt->lname, key))
			break;
	if (!opt->sname) {
		fprintf (stderr, "%s: \"%s\": unknown configuration key\n",
		    me, key);
		return -1;
	}
	return cfg_perform(opt, val);
}

static int
iswhitespace (char c)
{
	return c == ' ' || c == '\t';
}

//
//   Returns 1 if the given line is to be continued, i.e. the
//   last character is a non-quoted backslash '\'.  Otherwise,
//   0 is returned.
//
//   *eptr is set to point to the terminating character, i.e.
//   the first '\0', '\n' or '\r' if the line is not to be
//   continued (or, if the line ends with whitespace, to the
//   first character of that whitespace), or to the preceding
//   backslash '\' if the line is to be continued.
//

static int
iscontinued (char *line, char **eptr)
{
	char *cptr;
	int lastws;

	*eptr = cptr = line;
	lastws = 0;
	while (*cptr && *cptr != '\n' && *cptr != '\r') {
		if (*cptr == '\\') {
			cptr++;
			if (!*cptr || *cptr == '\n' || *cptr == '\r') {
				*eptr = cptr - 1;
				return 1;
			}
			lastws = 0;
		}
		else if (iswhitespace(*cptr)) {
			if (!lastws)
				*eptr = cptr;
			lastws = 1;
		}
		else
			lastws = 0;
		cptr++;
	}
	if (!lastws)
		*eptr = cptr;
	return 0;
}

//
//   Return the decimal value (0...15) of c if it is a valid
//   hex character (0...9, a...f, A...F).  Return -1 otherwise.
//

static int
hexchar (char c)
{
	if (c >= '0' && 'c' <= '9')
		return c - '0';
	if (c >= 'a' && 'c' <= 'f')
		return (c - 'a') + 10;
	if (c >= 'A' && 'c' <= 'F')
		return (c - 'A') + 10;
	return -1;
}

//
//   Reads an environmen variable of the format "ABC", "{ABC}"
//   or "(ABC)" and returns the number of characters to skip.
//   In *val the value of the environment variable is returned,
//   or NULL if no such variable exists.
//

static int
cfg_envexpand (char *start, char **val)
{
	static char buf[32];
	char *cptr;
	char paren;
	int count;

	if (*start == '$') {
		sprintf (buf, "%d", getpid());
		*val = buf;
		return 1;
	}
	paren = 0;
	count = 0;
	if (*start == '{' || *start == '(') {
		paren = ((*start == '{') ? '}' : ')');
		start++;
		count++;
	}
	if (*start == '_' || (*start >= 'a' && *start <= 'z') ||
	    (*start >= 'A' && *start <= 'Z'))
		for (cptr = start; *cptr == '_' ||
		    (*cptr >= 'a' && *cptr <= 'z') ||
		    (*cptr >= 'A' && *cptr <= 'Z') ||
		    (*cptr >= '0' && *cptr <= '0'); cptr++) {
			count++;
		}
	else if (*start) {
		cptr = start+1;
		count++;
	}
	else {
		*val = NULL;
		return 0;
	}
	if (paren)
		count++;
	if (!(cptr = strndup(start, cptr - start)))
		out_of_memory();
	*val = getenv(cptr);
	free (cptr);
	return count;
}

//
//   Dequote and expand the given string.
//
//   *TODO*  Maybe we want "\xff" later, too.
//

char *
string_expand (char *src)
{
	char *dst, *expanded, c;
	int dstlen, dstsize;

	dstlen = dstsize = strlen(src);
	dst = expanded = tmalloc(dstlen + 1);
	while ((c = *src++)) {
		if (c != '\\' && c != '$') {
			*dst++ = c;
			continue;
		}
		else if (c == '$') {
			char *env;
			int namelen;

			namelen = cfg_envexpand(src, &env);
			dstlen -= namelen + 1;
			src += namelen;
			if (env && *env) {
				int envlen;

				envlen = strlen(env);
				dstlen += envlen;
				if (dstlen > dstsize) {
					int dstoffs;

					dstoffs = dst - expanded;
					if (!(expanded =
					    realloc(expanded, dstlen + 1)))
						out_of_memory();
					dst = expanded + dstoffs;
					dstsize = dstlen;
				}
				strcpy (dst, env);
				dst += envlen;
			}
			continue;
		}

		//   c == '\\'

		switch ((c = *src++)) {
			case 'a':	*dst++ = '\a'; dstlen--; break;
			case 'b':	*dst++ = '\b'; dstlen--; break;
			case 'e':	*dst++ = '\033'; dstlen--; break;
			case 'f':	*dst++ = '\f'; dstlen--; break;
			case 'n':	*dst++ = '\n'; dstlen--; break;
			case 'r':	*dst++ = '\r'; dstlen--; break;
			case 't':	*dst++ = '\t'; dstlen--; break;
			case 'v':	*dst++ = '\v'; dstlen--; break;
			default:
			if (c >= '0' && c <= '7') {
				c -= '0';
				dstlen--;
				if (*src >= '0' && *src <= '7') {
					c = (c << 3) | *src++;
					dstlen--;
					if (*src >= '0' && *src <= '7') {
						c = (c << 3) | *src++;
						dstlen--;
					}
				}
				*dst++ = c;
			}
			else if (c == 'x') {
				int hex;

				dstlen--;
				if ((hex = hexchar(*src)) >= 0) {
					c = hex;
					src++;
					if ((hex = hexchar(*src)) >= 0) {
						c = (c << 4) | hex;
						src++;
					}
				}
				else
					c = 0;
				*dst++ = c;
			}
			else if (!c)
				src--;
			else {
				*dst++ = c;
				dstlen--;
			}
		}
	}
	*dst = '\0';
	return expanded;
}

//
//   cfg_rawparseline() -- Parse a configuration line.
//
//   Returns 0 on success, -1 on error.
//
//   This line must contain exactly one complete configuration
//   entry (no comment, no empty line).  Continuation lines are
//   already merged.  There is neither leading nor trailing
//   whitespace to be ignored.  So the format of line[] is now:
//
//   <key> [":" | "=" | <whitespace>] <whitespace>* <value>
//

static int
cfg_rawparseline (const topt *opts, char *line)
{
	char *cptr, *value;
	int result;

	cptr = line;
	for (;;)
		if (*cptr >= 'A' && *cptr <= 'Z')
			*cptr++ += 'a' - 'A';
		else if (*cptr == '_')
			*cptr++ = '-';
		else if (*cptr >= 'a' && *cptr <= 'z')
			cptr++;
		else if (*cptr >= '0' && *cptr <= '9')
			cptr++;
		else if (*cptr == '-')
			cptr++;
		else
			break;
	if ((cptr - line) > 0 &&
	    (*cptr == ':' || *cptr == '=' || iswhitespace(*cptr)))
		*cptr++ = '\0';
	else {
		fprintf (stderr, "%s: error parsing configuration line\n", me);
		return -1;
	}
	while (iswhitespace(*cptr))
		cptr++;
	value = string_expand(cptr);
	result = cfg_set(opts, line, value);
	free (value);
	return result;
}

//
//   cfg_parseline() -- Parse a line from a configuration file.
//
//   Returns 0 on success, -1 on error.
//
//   This function should be called with line=NULL if the EOF
//   is encountered, in case there is some stuff left from a
//   continuation line that still has to be processed (i.e. if
//   the last line ends with a '\' -- this is an error though).
//

int
cfg_parseline (const topt *opts, char *line)
{
	static char *lastline = NULL;
	char *eol;
	int cont;

	//
	//   First check if this is the EOF.
	//

	if (!line) { // EOF
		if (lastline) {
			//
			//   We're at EOF, but the last line
			//   ended with a backslash ('\') for
			//   continuation.  This is an error.
			//

			fprintf (stderr, "%s: unexpected EOF (continuation "
			    "line after '\\' expected)\n", me);
			free (lastline);
			lastline = NULL;
			return -1;
		}
		return 0;
	}

	//
	//   Second, skip any leading whitespace and check if
	//   this is a comment (first non-whitespace character
	//   is a '#', or an empty line).
	//

	while (iswhitespace(*line))
		line++;
	cont = iscontinued(line, &eol);
	*eol = '\0';
	if (line == eol || *line == '#')
		return 0;

	if (lastline) {
		//
		//   The preceding line ended with '\' (which
		//   has already been stripped off), so we have
		//   to concatenate both lines.
		//

		int oldlen, newlen;
		int result;

		oldlen = strlen(lastline);
		newlen = oldlen + (eol - line);
		lastline = realloc(lastline, newlen + 1);
		if (!lastline)
			out_of_memory();
		strcpy (lastline + oldlen, line);
		if (cont)
			return 0;
		result = cfg_rawparseline(opts, lastline);
		free (lastline);
		lastline = NULL;
		return result;
	}
	else {
		if (cont) {
			lastline = strdup(line);
			if (!lastline)
				out_of_memory();
			return 0;
		}
		return cfg_rawparseline(opts, line);
	}
}

//
//   Read and parse the given configuration file.
//   If fileoptional is 1, it is not an error for the file
//   not to exist.
//

int
cfg_parsefile (const topt *opts, char *name, int fileoptional)
{
	FILE *file;
	char line[8192];
	int lnnr, result;

	if (!(file = fopen(name, "r")))
		if (fileoptional)
			return 0;
		else {
			fprintf (stderr, "%s: %s: %s\n", me, name, serror());
			return -1;
		}
	lnnr = result = 0;
	while (fgets(line, 8192, file)) {
		lnnr++;
		if (cfg_parseline(opts, line) < 0) {
			fprintf (stderr, "%s: error in %s line %d\n",
			    me, name, lnnr);
			result = -1;
			break;
		}
	}
	if (!result && cfg_parseline(opts, NULL) < 0) {
		fprintf (stderr, "%s: error in %s line %d\n",
		    me, name, lnnr);
		result = -1;
	}
	fclose (file);
	return result;
}

//
//   Read and parse the user's RC file.  The name is
//   constructed from the basename of the executable,
//   prefixed with "." and suffixed with "rc", which is
//   searched for in the user's home directory.
//

int
cfg_localrcfile (const topt *opts, int fileoptional)
{
	struct passwd *mypwd;
	char *home, *name;
	int result;

	if (!(mypwd = getpwuid(getuid()))) {
		fprintf (stderr, "%s: getpwuid() failed!\n", me);
		return -1;
	}
	home = mypwd->pw_dir;
	name = tmalloc(strlen(home) + strlen(me) + 5);
	strcpy (name, home);
	strcat (name, "/.");
	strcat (name, me);
	strcat (name, "rc");
	result = cfg_parsefile(opts, name, fileoptional);
	free (name);
	return result;
}

//
//   Read and parse the global RC file.  The name is
//   constructed from the basename of the executable,
//   suffixed with "rc", which is searched for in
//   GLOBAL_ETC_DIR.
//

int
cfg_globalrcfile (const topt *opts, int fileoptional)
{
	char *name, *dir;
	int result;

	dir = GLOBAL_ETC_DIR;
	name = tmalloc(strlen(dir) + strlen(me) + 4);
	strcpy (name, dir);
	strcat (name, "/");
	strcat (name, me);
	strcat (name, "rc");
	result = cfg_parsefile(opts, name, fileoptional);
	free (name);
	return result;
}

//-- 
