//
//   ftp.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: ftp.c,v 1.35 1998/12/12 22:33:57 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: ftp.c,v 1.35 1998/12/12 22:33:57 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <errno.h>
#include <limits.h>

#include "utils.h"
#include "omi.h"
#include "ftp.h"
#include "pro.h"

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

static volatile sig_atomic_t have_alarm;

static void
catch_alarm (int signum)
{
	have_alarm = 1;
}

void
set_alarm (unsigned seconds)
{
	have_alarm = 0;
	alarm (seconds);
}

void
unset_alarm (void)
{
	alarm (0);
}

//
//   Initialize an ftpstat structure.  Must be done before
//   using it in ftp_rawopen().
//

ftpstat *
ftp_initserver (char *hostspec, unsigned int port)
{
	ftpstat *fs;

	alarm (0);
	catchsignal (SIGALRM, catch_alarm);
	signal (SIGPIPE, SIG_IGN);
	fs = tmalloc(sizeof(*fs));
	if (!(fs->name = strdup(hostspec)))
		out_of_memory();
	fs->ip.s_addr = INADDR_NONE;
	fs->lip.s_addr = INADDR_NONE;
	fs->port = port;
	fs->fd = -1;
	fs->cwd = NULL;
	fs->input = NULL;
	fs->inputsize = 0;
	fs->type = TYPE_UNKNOWN;
	fs->support = 0xffff;
	fs->hretry = 0;
	fs->sretry = 0;
	fs->remfail = 0;
	fs->locfail = 0;
	fs->rodelay = cc->reopendelay;
	fs->pasvfail = 0;
	fs->pasvsuccess = 0;
	fs->pasvlast = 0;
	return fs;
}

//
//   Simple wrapper function for write().
//   It simply blocks until all the stuff is transferred,
//   or until a SIGALRM is delivered, which means that a
//   timeout occured.  Returns the number of bytes actually
//   written -- if it differs from the specified "count",
//   an error or timeout occured.
//

static int
ftp_write (ftpstat *fs, char *data, int count)
{
	int byteswritten, result;

	byteswritten = 0;
	while (count) {
		result = write(fs->fd, data, count);
		if (result < 0 && errno != EINTR) {
			fprintf (stderr, "%s: %s: write(): %s\n",
			    me, fs->name, serror());
			if (errno == EPIPE) {	// kaputte Pfeife
				fprintf (stderr, "%s: %s: write(): server "
				    "closed the control connection "
				    "unexpectedly\n", me, fs->name);
				close (fs->fd);
				fs->fd = -1;
			}
			break;
		}
		if (result > 0) {
			count -= result;
			data += result;
			byteswritten += result;
		}
		if (have_alarm) {
			fprintf (stderr, "%s: %s: write() timeout\n",
			    me, fs->name);
			break;
		}
	}
	return byteswritten;
}

//
//   Wrapper function for read().
//   It reads the reply from an FTP control connection, which
//   may be multiple lines.  Returns 0 on error or timeout.
//   *data is set to the complete reply, while *line points
//   to the 5th character of the last line, i.e. the string
//   after the result code (which is the substring of *data
//   that usually is the most "interesting" part).
//

static unsigned int
ftp_read (ftpstat *fs, char **data, char **line)
{
	fd_set rfdset;
	int result, lnoffs; /* NOT used uninitialized (ignore stupid gcc) */
	int bytesread, check, multiline;
	char buff[8192], *cptr, *newln;

	bytesread = 0;
	*data = NULL;
	if (fs->input) {
		multiline = -1;
		result = fs->inputsize;
		memcpy (buff, fs->input, result);
		fs->input = NULL;
	}
	else
		multiline = 0;
	for (;;) {
		if (multiline == -1)
			multiline = 0;
		else {
			FD_ZERO (&rfdset);
			FD_SET (fs->fd, &rfdset);
			set_alarm (cc->timeout_read);
			result = select(FD_SETSIZE, &rfdset, NULL, NULL, NULL);
			if (result >= 0)
				if (result > 0 && FD_ISSET(fs->fd, &rfdset)) {
					result = read(fs->fd, buff, 8192);
					if (!result) {
						unset_alarm();
						fprintf (stderr, "%s: %s: "
						    "read(): server closed "
						    "the control connection "
						    "unexpectedly\n",
						    me, fs->name);
						close (fs->fd);
						fs->fd = -1;
						if (*data)
							free (*data);
						return 0;
					}
				}
				else
					result = 0;
			unset_alarm();
			if (result < 0 && errno != EINTR) {
				fprintf (stderr, "%s: %s: read(): %s\n",
				    me, fs->name, serror());
				if (*data)
					free (*data);
				return 0;
			}
			if (have_alarm) {
				fprintf (stderr, "%s: %s: read() timeout\n",
				    me, fs->name);
				if (*data)
					free (*data);
				return 0;
			}
		}
		if (result > 0) {
			if (!(*data = realloc(*data, bytesread + result)))
				out_of_memory();
			memcpy (*data + bytesread, buff, result);
			bytesread += result;
			if (multiline == 0) {
				//
				//   We don't know yet whether
				//   this is a single-line or a
				//   multi-line reply.  We need
				//   at least the first 4 chars
				//   to decide about this.
				//   See RFC 959 for details.
				//

				if (bytesread >= 4)
					if ((*data)[3] == '-')
						multiline = 2;
					else {
						multiline = 1;
						lnoffs = 4;
					}
			}
			if (multiline == 1) {
				//
				//   This is a single line
				//   reply.  We read until the
				//   first '\n'.
				//

				if ((newln = memchr(*data, '\n', bytesread)))
					break;
			}
			else if (multiline == 2) {
				//
				//   This is a multiple line
				//   reply, and we haven't
				//   encountered the last line
				//   yet.  We are looking for
				//   the beginning of the last
				//   line, i.e. '\n' followed
				//   by 3 digits and a space.
				//

				if ((check = result + 4) > bytesread)
					check = bytesread;
				cptr = (*data) + (bytesread - check);
				if ((check -= 4) <= 0)
					check = 1;
				while ((newln = memchr(cptr, '\n', check))) {
					if (newln[1] >= '0' &&
					    newln[1] <= '9' &&
					    newln[2] >= '0' &&
					    newln[2] <= '9' &&
					    newln[3] >= '0' &&
					    newln[3] <= '9' &&
					    newln[4] == ' ') {
						multiline = 4;
						lnoffs = (newln - *data) + 5;
						break;
					}
					else {
						check -= (newln - cptr) + 1;
						if (check <= 0)
							break;
						cptr = newln + 1;
					}
				}
			} /* if (multiline == 2) */
			if (multiline == 4) {
				//
				//   This is a multiple line
				//   reply, and we've already
				//   encountered the last line.
				//   We read until the '\n'
				//   that terminates that last
				//   line.
				//

				if (bytesread > lnoffs &&
				    (newln = memchr((*data) + lnoffs,
				    '\n', bytesread - lnoffs)))
					break;
			}
		} /* if (result > 0) */
	} /* for (;;) */
	*line = *data + lnoffs;

	//
	//   The last '\n' is replaced with a terminating zero.
	//   If it is preceded by a '\r' (as it should be
	//   according to the telnet protocol standard), that
	//   one is also replaced with a zero.
	//

	*newln = '\0';
	if (newln[-1] == '\r')
		newln[-1] = '\0';
	check = (newln - *data) + 1;
	if (bytesread > check) {
		fs->inputsize = bytesread - check;
		fs->input = tmalloc(fs->inputsize);
		memcpy (fs->input, newln + 1, fs->inputsize);
		*data = realloc(*data, check);
	}
	return check;
}

//
//   "Fake" an ftpreply structure with the given reply code
//   and text.
//

ftpreply *
ftp_fakereply (int code, int retry, char *format, /* arguments */ ...)
{
	va_list ap;
	ftpreply *fr;
	int txtlen;
#ifdef HAVE_VSNPRINTF
	char dummy;
#endif

	fr = tmalloc(sizeof(*fr));
	fr->code = code;
	fr->retry = retry;
#ifdef HAVE_VSNPRINTF
	va_start (ap, format);
	txtlen = vsnprintf(&dummy, 1, format, ap) + 5;
	va_end (ap);
#else
	//
	//   Shit, we don't have a working vsnprintf().
	//   This is a quick 'n dirty work-around.  :-(
	//
	txtlen = PATH_MAX + strlen(format) + 100;
#endif
	fr->text = tmalloc(txtlen);
	sprintf (fr->text, "%03d ", code);
	fr->line = fr->text + 4;
	va_start (ap, format);
	if (vsprintf(fr->line, format, ap) > txtlen - 5) {
		fprintf (stderr,
		    "%s: Unexpected buffer overflow in vsprintf()!\n", me);
		abort();
	}
	va_end (ap);
	if (code >= 400)
		fprintf (stderr, "%s: %s\n", me, fr->line);
	return fr;
}

//
//   Reads a reply from the remote FTP server and builds an
//   appropriate ftpreply structure.
//

static ftpreply *
ftp_readreply (ftpstat *fs)
{
	ftpreply *fr;

	fr = tmalloc(sizeof(*fr));
	if (!ftp_read(fs, &fr->text, &fr->line)) {
		free (fr);
		return ftp_fakereply(499, RETRY_REOPEN,
		    "read() from control connection failed");
	}
	fr->code = (fr->text[0] - '0') * 100
	         + (fr->text[1] - '0') * 10
	         + (fr->text[2] - '0');
	if (cc->debug & DEBUG_INPUT)
		if (cc->debug & DEBUG_MULTI)
			fprintf (stderr, "--> %s\n", fr->text);
		else
			fprintf (stderr, "--> %d %s\n", fr->code, fr->line);
	fr->retry = RETRY_NO;
	return fr;
}

//
//   Frees the given ftpreply structure.
//

void
ftp_freereply (ftpreply *fr)
{
	if (fr) {
		if (fr->text)
			free (fr->text);
		free (fr);
	}
}

ftpreply *
ftp_cmd (ftpstat *fs, char *format, /* arguments */ ...)
{
	va_list ap;
	int cmdlen, result;
	char *cmd;
#ifdef HAVE_VSNPRINTF
	char dummy;
#endif

	//
	//   First we build the command string from the
	//   given format and arguments.
	//

#ifdef HAVE_VSNPRINTF
	va_start (ap, format);
	cmdlen = vsnprintf(&dummy, 1, format, ap) + 2;
	va_end (ap);
#else
	//
	//   Shit, we don't have a working vsnprintf().
	//   This is a quick 'n dirty work-around.  :-(
	//
	cmdlen = PATH_MAX + strlen(format) + 100;
#endif
	cmd = tmalloc(cmdlen + 1);
	va_start (ap, format);
	if ((result = vsprintf(cmd, format, ap) + 2) > cmdlen) {
		fprintf (stderr,
		    "%s: Unexpected buffer overflow in vsprintf()!\n", me);
		abort();
	}
	cmdlen = result;
	va_end (ap);
	if (cc->debug & DEBUG_OUTPUT)
		fprintf (stderr, "<-- %s\n", cmd);
	strcat (cmd, "\r\n");

	//
	//   Now send the command to the server, paying
	//   attention to the command timeout value.
	//

	set_alarm (cc->timeout_command);
	result = ftp_write(fs, cmd, cmdlen);
	unset_alarm();
	if (result != cmdlen)
		return ftp_fakereply(499, RETRY_REOPEN,
		    "write() to control connection failed");
	free (cmd);

	//
	//   Finally read the reply from the server and
	//   return it.
	//

	return ftp_readreply(fs);
}

//
//   Close the control connection.
//

static ftpreply *
ftp_rawclose (ftpstat *fs)
{
	ftpreply *fr;

	if (fs->fd >= 0) {
		//
		//   We send a "QUIT" just to make the server
		//   happy, but we close the connection, no
		//   matter what happens.
		//
		//   Note that a timeout or some other trouble
		//   might have occured before ftp_rawclose()
		//   has been called, so the result might be
		//   meaningless.
		//

		fr = ftp_cmd(fs, "QUIT");
		close (fs->fd);
		fs->fd = -1;
	}
	else {
		//
		//   The control connection has already been
		//   closed.
		//

		fr = ftp_fakereply(299, RETRY_NO,
		    "control connection closed");
	}
	if (cc->debug & DEBUG_CONNECT)
		fprintf (stderr, "control connection closed\n");
	return fr;
}

//
//   Close the control connection (if necessary) and
//   free the ftpstat structure.
//

ftpreply *
ftp_doneserver (ftpstat *fs)
{
	ftpreply *fr;

	fr = ftp_rawclose(fs);
	if (fs->name)
		free (fs->name);
	if (fs->input)
		free (fs->input);
	if (fs->cwd)
		free (fs->cwd);
	free (fs);
	return fr;
}

//
//   Perform DNS look-up.
//

static ftpreply *
get_ip (char *hostspec, struct in_addr *ip)
{
	char *cptr;
	int isip;

	//
	//   First check whether this is an IP number or a
	//   host name, in the latter case perform a DNS
	//   look-up to resolve the name.  We need to make
	//   sure that the look-up doesn't block longer than
	//   the corresponding timeout value.
	//

	isip = 1;
	for (cptr = hostspec; *cptr; cptr++)
		if ((*cptr < '0' || *cptr > '9') && *cptr != '.') {
			isip = 0;
			break;
		}
	if (isip) {
		//
		//   It's an IP number.  Convert the string to
		//   an integer in network byte order.
		//
		if ((ip->s_addr = inet_addr(hostspec)) == INADDR_NONE)
			return ftp_fakereply(599, RETRY_NO,
			    "%s: malformed IP number", hostspec);
	}
	else {
		struct hostent *myhostent;

		if (cc->debug & DEBUG_DNS)
			fprintf (stderr, "looking up hostname\n");
		set_alarm (cc->timeout_dns);
		myhostent = gethostbyname(hostspec);
		unset_alarm();
		if (!myhostent) {
			if (have_alarm)
				fprintf (stderr,
				    "%s: %s: DNS lookup timeout\n",
				    me, hostspec);
			return ftp_fakereply(499,
			    have_alarm ? RETRY_REOPEN : RETRY_NO,
			    "%s: can't resolve hostname", hostspec);
		}
		memcpy (&ip->s_addr, myhostent->h_addr, sizeof(ip->s_addr));
	}
	return ftp_fakereply(299, RETRY_NO, "DNS lookup: %s -> %s",
	    hostspec, inet_ntoa(*ip));
}

//
//   Open the actual TCP connection to the remote FTP server
//   and login using the USER and PASS commands.
//

static ftpreply *
ftp_rawopen (ftpstat *fs)
{
	ftpreply *fr;
	struct sockaddr_in server, client;
	int result;

	//
	//   First check whether we have to do a host name
	//   look-up.

	if (fs->ip.s_addr == INADDR_NONE) {
		fr = get_ip(fs->name, &fs->ip);
		if (fr->code >= 400)
			return fr;
		if (cc->debug & DEBUG_DNS)
			fprintf (stderr, "%s\n", fr->line);
		ftp_freereply (fr);
	}

	//
	//   Clean up the ftpstat structure.
	//

	if (fs->input) {
		free (fs->input);
		fs->input = NULL;
	}
	if (fs->cwd) {
		free (fs->cwd);
		fs->cwd = NULL;
	}
	fs->type = TYPE_UNKNOWN;

	//
	//   Get a file descriptor for a TCP socket.
	//

	if ((fs->fd = socket(AF_INET, SOCK_STREAM, 6)) < 0)
		return ftp_fakereply(499, RETRY_REOPEN,
		    "socket(): %s", serror());

	//
	//   Bind the socket to a local address,
	//   if we have been configured to do so.
	//

	if (cc->local_addr) {
		if (cc->debug & DEBUG_CONNECT)
			fprintf (stderr,
			    "binding local socket to %s\n", cc->local_addr);
		memset (&client, 0, sizeof(client));
		client.sin_family = AF_INET;

		if (fs->lip.s_addr == INADDR_NONE) {
			fr = get_ip(cc->local_addr, &fs->lip);
			if (fr->code >= 400) {
				close (fs->fd);
				fs->fd = -1;
				return fr;
			}
			if (cc->debug & DEBUG_DNS)
				fprintf (stderr, "%s\n", fr->line);
			ftp_freereply (fr);
		}

		client.sin_addr.s_addr = fs->lip.s_addr;
		result = bind(fs->fd,
		    (struct sockaddr *)&client, sizeof(client));
		if (result < 0) {
			close (fs->fd);
			fs->fd = -1;
			return ftp_fakereply(499, RETRY_REOPEN,
			    "binding control connection to %s: %s",
			    inet_ntoa(fs->lip), serror());
		}
		if (cc->debug & DEBUG_CONNECT)
			fprintf (stderr,
			    "local socket bound to %s\n", inet_ntoa(fs->lip));
	}

	//
	//   Try to open a TCP connection, paying
	//   attention to the connection timeout value.
	//

	if (cc->debug & DEBUG_CONNECT)
		fprintf (stderr,
		    "opening control connection to %s on port %u\n",
		    inet_ntoa(fs->ip), fs->port);
	memset (&server, 0, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(fs->port);
	server.sin_addr.s_addr = fs->ip.s_addr;
	set_alarm (cc->timeout_connect);
	result = connect(fs->fd, (struct sockaddr *)&server, sizeof(server));
	unset_alarm();
	if (result < 0) {
		close (fs->fd);
		fs->fd = -1;
		if (have_alarm)
			return ftp_fakereply(499, RETRY_REOPEN,
			    "opening control connection to %s: "
			    "connect() timeout", fs->name);
		else
			return ftp_fakereply(499, RETRY_REOPEN,
			    "opening control connection to %s: %s",
			    fs->name, serror());
	}
	if (cc->debug & DEBUG_CONNECT)
		fprintf (stderr, "control connection opened\n");

	//
	//   Wait for the "welcome" string sent by the server.
	//

	fr = ftp_readreply(fs);
	if (fr->code >= 400) {
		fr->retry = RETRY_REOPEN;
		return fr;
	}
	if (fr->code <= 199) {
		fr->code = 499;	// make it a temporary error
		fr->retry = RETRY_REOPEN;
		return fr;
	}
	ftp_freereply (fr);

	//
	//   Send USER and PASS commands.
	//

	fr = ftp_cmd(fs, "USER %s", cc->login);
	if (fr->code >= 400) {
		if (fr->code == 530) {
			//
			//   This is probably "too many users",
			//   so make it a temporary error.
			//
			fr->code = 499;
			fr->retry = RETRY_REOPEN;
		}
		return fr;
	}
	if (fr->code == 331) {
		ftp_freereply (fr);
		fr = ftp_cmd(fs, "PASS %s", cc->email);
		if (fr->code >= 400) {
			if (fr->code == 530) {
				//   See above.
				fr->code = 499;
				fr->retry = RETRY_REOPEN;
			}
			return fr;
		}
	}
	if (fr->code == 332) {
		ftp_freereply (fr);
		if (!cc->account)
			return ftp_fakereply(599, RETRY_NO,
			    "Server requires account information for login");
		fr = ftp_cmd(fs, "ACCT %s", cc->account);
		if (fr->code >= 400) {
			if (fr->code == 530) {
				//   See above.
				fr->code = 499;
				fr->retry = RETRY_REOPEN;
			}
			return fr;
		}
	}
	if (fr->code >= 300) {
		ftp_freereply (fr);
		return ftp_fakereply(599, RETRY_NO,
		    "error in login sequence (server violates RFC959)");
	}
	ftp_freereply (fr);
	return ftp_fakereply(299, RETRY_NO, "control connection established");
}

//
//   Run a sequence of FTP commands, taking care of retries
//   if something goes wrong.
//

ftpreply *
ftp_sequence (ftpstat *fs, sequencejob *job, void *data)
{
	ftpreply *fr;
	int retry;
	bool done;

	retry = RETRY_NO;
	done = FALSE;
	fr = NULL;
	fs->sretry = 0;
	fs->rodelay = cc->reopendelay;
	do {
		if (fr) {
			ftp_freereply (fr);
			fr = NULL;
		}

		//
		//   If this is a "reopen" retry, close the control
		//   connection (ignore the result) and sleep a bit.
		//

		if (retry == RETRY_REOPEN) {
			ftp_freereply (ftp_rawclose(fs));
			if (fs->rodelay) {
				if (cc->debug & DEBUG_CONNECT)
					fprintf (stderr,
					    "sleeping %d seconds "
					    "before retrying\n",
					    fs->rodelay);
				sleep (fs->rodelay);
			}
			if ((fs->rodelay += cc->reopeninc) > cc->reopenmax)
				fs->rodelay = cc->reopenmax;
		}
		retry = RETRY_NO;

		//
		//   First check whether the control connection
		//   has to be opened.
		//

		if (fs->fd < 0) {
			fr = ftp_rawopen(fs);
			if (fr->code <= 399) {
				ftp_freereply (fr);
				fr = NULL;
			}
		}

		//
		//   Now execute the requested sequence of FTP
		//   commands.
		//

		if (!fr)
			fr = job(fs, data);
		if (fr->code >= 400) {
			if (fs->locfail >= cc->maxlocfail)
				fr->retry = RETRY_STOP;
			retry = fr->retry;
		}

		if (retry == RETRY_NO || retry == RETRY_STOP)
			done = TRUE;
		else if (retry == RETRY_RETRY && fs->sretry++ < cc->maxsretry)
			done = FALSE;
		else {
			fs->sretry = 0;
			if (fs->hretry++ < cc->maxhretry) {
				retry = RETRY_REOPEN;
				done = FALSE;
			}
			else {
				if (++fs->remfail >= cc->maxremfail)
					fr->retry = RETRY_STOP;
				done = TRUE;
			}
		}

	} while (!done);
	return fr;
}

//
//   Set the transfer type (ASCII / image), if necessary.
//

ftpreply *
ftp_needtype (ftpstat *fs, ftpxfertype reqtype)
{
	ftpreply *fr;

	if (fs->type != reqtype) {
		fr = ftp_cmd(fs, "TYPE %c", reqtype);
		if (fr->code >= 200 && fr->code <= 299)
			fs->type = reqtype;
		else {
			if (fr->code <= 399) {
				fr->code = 499;	// make it an temporary error
				fr->retry = RETRY_RETRY;
			}
			fs->type = TYPE_UNKNOWN;
		}
	}
	else
		fr = ftp_fakereply(299, RETRY_NO, "type set to %c", fs->type);
	return fr;
}

//
//   Change the current working directory, if necessary.
//

ftpreply *
ftp_needcwd (ftpstat *fs, char *reqcwd)
{
	ftpreply *fr;

	if (!fs->cwd || strcmp(fs->cwd, reqcwd)) {
		fr = ftp_cmd(fs, "CWD %s", reqcwd);
		if (fs->cwd)
			free (fs->cwd);
		if (fr->code >= 200 && fr->code <= 299) {
			if (!(fs->cwd = strdup(reqcwd)))
				out_of_memory();
		}
		else {
			if (fr->code <= 399) {
				fr->code = 499;	// make it a temporary error
				fr->retry = RETRY_RETRY;
			}
			fs->cwd = NULL;
		}
	}
	else
		fr = ftp_fakereply(299, RETRY_NO, "cwd set to %s", fs->cwd);
	return fr;
}

//
//   Try to parse a "227" reply (in response to a PASV
//   command) and extract the remote server's IP and
//   port number.  Returns 0 on success, -1 on error.
//

static int
parse_pasvreply (char *line, struct in_addr *ip, unsigned short *port)
{
	char *cptr, *eptr;
	int num[6];
	int numnums;

	cptr = line;
	numnums = 0;
	while (numnums < 6) {
		while (*cptr < '0' || *cptr > '9')
			if (!*cptr++)
				return -1;
		num[numnums] = strtol(cptr, &eptr, 10);
		cptr = eptr;
		if ((*cptr != ',' && numnums < 5)
		    || num[numnums] < 0 || num[numnums] > 255) {
			numnums = 0;
			continue;
		}
		cptr++;
		if ((*cptr < '0' || *cptr > '9') && numnums < 5) {
			numnums = 0;
			continue;
		}
		numnums++;
	}
	ip->s_addr = htonl((unsigned long) num[0] << 24
	    | (unsigned long) num[1] << 16
	    | (unsigned long) num[2] << 8
	    | (unsigned long) num[3]);
	*port = htons((unsigned short) num[4] << 8 | (unsigned short) num[5]);
	return 0;
}

//
//   Copies data from one descriptor to the other,
//   until either EOF or an error occurs.
//   Return values:
//      >= 0  success (returns number of bytes transferred)
//        -1  read error (network)
//        -2  write error (local disk)
//

#define BUFFSIZE 65536

static long
fdcopy (int src, int dst)
{
	//
	// *TODO* The buffer should not be a local variable.
	//        Instead we should malloc() it.  (Should we?)
	//

	fd_set rfdset, wfdset;
	char buff[BUFFSIZE];	//  ring buffer
	int result, bufstart, bufend, buffill, done;
	unsigned long size;

	done = FALSE;
	bufstart = bufend = buffill = 0;
	size = 0;
	for (;;) {
		FD_ZERO (&rfdset);
		FD_ZERO (&wfdset);
		if (buffill < BUFFSIZE && !done)
			FD_SET (src, &rfdset);
		if (buffill > 0)
			FD_SET (dst, &wfdset);
		else
			if (done)
				return size;
		set_alarm (cc->timeout_read);
		result = select(FD_SETSIZE, &rfdset, &wfdset, NULL, NULL);
		if (result < 0) {
			unset_alarm();
			if (have_alarm) {
				fprintf (stderr, "%s: data connection: "
				    "read timeout (select())\n",
				    me);
				return -1;
			}
			if (errno == EINTR)
				continue;
			fprintf (stderr,
			    "%s: data connection: select(): %s\n",
			    me, serror());
			return -1;
		}
		if (result > 0 && FD_ISSET(src, &rfdset)) {
			if (bufend >= bufstart)
				result = read(src, buff + bufend,
				    BUFFSIZE - bufend);
			else
				result = read(src, buff + bufend,
				    bufstart - bufend);
			unset_alarm();
			if (!result)
				done = TRUE;
			if (result < 0) {
				if (errno == EINTR)
					continue;
				fprintf (stderr,
				    "%s: data connection: read(): %s\n",
				    me, serror());
				return -1;
			}
			if (have_alarm) {
				fprintf (stderr,
				    "%s: data connection: read() timeout\n",
				    me);
				return -1;
			}
			buffill += result;
			if ((bufend += result) >= BUFFSIZE)
				bufend -= BUFFSIZE;
		}
		else if (result > 0 && FD_ISSET(dst, &wfdset)) {
			unset_alarm();
			if (bufend > bufstart)
				result = write(dst, buff + bufstart,
				    bufend - bufstart);
			else
				result = write(dst, buff + bufstart,
				    BUFFSIZE - bufstart);
			if (result < 0) {
				if (errno == EINTR)
					continue;
				fprintf (stderr,
				    "%s: local file: write(): %s\n",
				    me, serror());
				return -2;
			}
			size += result;
			update_progress (size);
			buffill -= result;
			if ((bufstart += result) >= BUFFSIZE)
				bufstart -= BUFFSIZE;
		}
		else
			unset_alarm();
	}
}

//
//   Copies data from one descriptor to memory,
//   until either EOF or an error occurs.
//   Returns values:
//      >=0 success (returns number of bytes read, without terminating zero)
//      -1  read error (network)
//      -2  write error (out of memory)
//

static long
lscopy (int src, char **dstptr)
{
	//
	// *TODO* The buffer should not be a local variable.
	//        Instead we should malloc() it.  (Should we?)
	//

	fd_set rfdset;
	char buff[BUFFSIZE];
	char *dst, *new;
	unsigned long size;	//   Without terminating zero!
	int result, buffill, done;

	dst = NULL;
	done = 0;
	buffill = 0;
	size = 0;
	for (;;) {
		FD_ZERO (&rfdset);
		FD_SET (src, &rfdset);
		set_alarm (cc->timeout_read);
		result = select(FD_SETSIZE, &rfdset, NULL, NULL, NULL);
		if (result >= 0)
			if (result > 0 && FD_ISSET(src, &rfdset))
				result = read(src, buff + buffill,
				    BUFFSIZE - buffill);
			else
				//   This shouldn't happen!
				continue;
		unset_alarm();
		if (result < 0)
			if (errno != EINTR) {
				fprintf (stderr,
				    "%s: data connection: read(): %s\n",
				    me, serror());
				if (dst)
					free (dst);
				*dstptr = NULL;
				return -1;
			}
			else if (!have_alarm)
				continue;
		if (have_alarm) {
			fprintf (stderr,
			    "%s: data connection: read() timeout\n", me);
			if (dst)
				free (dst);
			*dstptr = NULL;
			return -1;
		}
		if ((buffill += result) == BUFFSIZE || (done = !result)) {
			if (buffill) {
				if (!(new = realloc(dst, size + buffill + 1))) {
					fprintf (stderr,
					    "%s: realloc() failed\n", me);
					if (dst)
						free (dst);
					*dstptr = NULL;
					return -2;
				}
				dst = new;
				memcpy (dst + size, buff, buffill);
				size += buffill;
				update_progress (size);
				buffill = 0;
			}
			if (done) {
				if (!dst) {
					if (!(dst = malloc(1))) {
						fprintf (stderr,
						    "%s: malloc() failed\n",
						    me);
						*dstptr = NULL;
						return -2;
					}
				}
				*dstptr = dst;
				dst[size] = 0;	//  Terminate with a zero.
				return size;
			}
		}
	}
}

static ftpreply *
pre_initconnection_passive (ftpstat *fs, int *socke)
{
	//
	//   Prepare the data connection for passive transfer.
	//   This is performed BEFORE the transfer command is
	//   sent to the server.
	//

	ftpreply *fr;
	struct sockaddr_in data;
	int result;

	//
	//   Enter passive mode, and parse the server's reply
	//   for the IP/port of the data connection.
	//

	fr = ftp_cmd(fs, "PASV");
	if (fr->code != 227) {
		//
		//   This server doesn't seem to support
		//   passive FTP.  Give up for now...
		//
		if (!(cc->debug & DEBUG_INPUT))
			fprintf (stderr, "PASV: %d %s\n", fr->code, fr->line);
		fprintf (stderr, "unexpected reply code %d to PASV command\n",
		    fr->code);
		fs->pasvfail++;
		fs->pasvlast++;
		if (fs->pasvfail <= 10 || fs->pasvsuccess >= fs->pasvfail
		    || fs->pasvlast <= 10) {
			if (fr->code <= 399)
				fr->code = 499;
			fr->retry = RETRY_RETRY;
		}
		else {
			if (fr->code <= 499)
				fr->code = 599;	// make it a permanent error
			if (fs->pasvsuccess > 0) {
				fprintf (stderr, "Excessive number of PASV "
				    "failures, re-opening connection.\n");
				fs->pasvfail = 0;
				fs->pasvsuccess = 0;
				fs->pasvlast = 0;
				fr->retry = RETRY_REOPEN;
			}
			else {
				fprintf (stderr, "PASV failed too many "
				    "times for this server, giving up.\n");
				fr->retry = RETRY_STOP;
			}
		}
		return fr;
	}
	result = parse_pasvreply(fr->line, &data.sin_addr, &data.sin_port);
	ftp_freereply (fr);
	if (result < 0) {
		//
		//   Can't recognize IP and port numbers in
		//   the remote PASV reply.
		//
		return ftp_fakereply(599, RETRY_STOP,
		    "can't parse IP and port number in reply to "
		    "PASV command (227)");
	}
	fs->pasvsuccess++;
	fs->pasvlast = 0;

	//
	//   Try to open a data connection to the FTP server.
	//   First get a file descriptor for a TCP socket.
	//

	if ((*socke = socket(AF_INET, SOCK_STREAM, 6)) < 0)
		return ftp_fakereply(499, RETRY_RETRY,
		    "socket(): %s", serror());

	//
	//   Try to open a TCP connection on the port reported
	//   by the FTP server in reply to the PASV command,
	//   and pay attention to the connection timeout value.
	//

	if (cc->debug & DEBUG_CONNECT) {
		struct in_addr ia;

		memcpy (&ia, &data.sin_addr, sizeof(ia));
		fprintf (stderr, "opening data connection to %s on port %u\n",
		    inet_ntoa(ia), ntohs(data.sin_port));
	}
	data.sin_family = AF_INET;
	set_alarm (cc->timeout_connect);
	result = connect(*socke, (struct sockaddr *) &data, sizeof(data));
	unset_alarm();
	if (result < 0) {
		close (*socke);
		if (have_alarm)
			return ftp_fakereply(499, RETRY_RETRY,
			    "opening data connection to %s: "
			    "connect() timeout", fs->name);
		else
			return ftp_fakereply(499, RETRY_RETRY,
			    "opening data connection to %s: %s",
			    fs->name, serror());
	}

	return ftp_fakereply(299, RETRY_NO,
	    "pre_initconnection_passive() successful");
}

static ftpreply *
pre_initconnection_active (ftpstat *fs, int *socke)
{
	//
	//   Prepare the data connection for active transfer.
	//   This is performed BEFORE the transfer command is
	//   sent to the server.
	//

	ftpreply *fr;
	struct sockaddr_in datalisten;
	unsigned char *aptr, *pptr;
	int llen;

	//
	//   Try to open a socket for listening to the incoming
	//   data connection from the FTP server.
	//   First get a file descriptor for a TCP socket.
	//

	if ((*socke = socket(AF_INET, SOCK_STREAM, 6)) < 0)
		return ftp_fakereply(499, RETRY_RETRY, "socket(): %s",
		    serror());

	//
	//   Bind a local address to our socket for listening.
	//   Use the properties of the local part of the control
	//   connection.  Setting the port number to zero lets
	//   bind() choose an appropriate port number.
	//

	llen = sizeof(datalisten);
	if (getsockname(fs->fd, (struct sockaddr *) &datalisten, &llen) < 0) {
		close (*socke);
		return ftp_fakereply(499, RETRY_RETRY, "getsockname(): %s",
		    serror());
	}
	datalisten.sin_port = 0;
	llen = sizeof(datalisten);
	if (bind(*socke, (struct sockaddr *) &datalisten, llen) < 0) {
		close (*socke);
		return ftp_fakereply(499, RETRY_RETRY, "bind(): %s",
		    serror());
	}

	//
	//   Find out which port number was assigned to our
	//   socket by bind().
	//

	if (getsockname(*socke, (struct sockaddr *) &datalisten, &llen) < 0) {
		close (*socke);
		return ftp_fakereply(499, RETRY_RETRY, "getsockname(): %s",
		    serror());
	}

	//
	//   Listen for exactly one incoming data connection.
	//

	if (listen(*socke, 1) < 0) {
		close (*socke);
		return ftp_fakereply(499, RETRY_RETRY, "listen(): %s",
		    serror());
	}

	//
	//   Send a PORT command to the server, telling him
	//   about our address and port number to connect to.
	//
	//   The extraction of the bytes looks messy, but it
	//   actually works fine because address and port are
	//   in network byte order.
	//

	aptr = (unsigned char *) &datalisten.sin_addr;
	pptr = (unsigned char *) &datalisten.sin_port;
	fr = ftp_cmd(fs, "PORT %u,%u,%u,%u,%u,%u",
	    aptr[0], aptr[1], aptr[2], aptr[3], pptr[0], pptr[1]);
	if (fr->code >= 400) {
		close (*socke);
		return fr;
	}

	return ftp_fakereply(299, RETRY_NO,
	    "pre_initconnection_active() successful");
}

static ftpreply *
post_initconnection_active (ftpstat *fs, int *socke)
{
	//
	//   Prepare the data connection for active transfer.
	//   This is performed AFTER the transfer command has
	//   been sent to the server.
	//

	struct sockaddr_in datasrv;
	int srvlen, datasocke;

	//
	//   Wait for the server's data connection and
	//   accept it.  After that, the listening socket
	//   isn't needed anymore, so we close it and
	//   replace it by the connection socket.
	//

	srvlen = sizeof(datasrv);
	if ((datasocke = accept(*socke, (struct sockaddr *) &datasrv, &srvlen))
	    < 0) {
		close (*socke);
		return ftp_fakereply(499, RETRY_RETRY, "accept(): %s",
		    serror());
	}
	close (*socke);
	*socke = datasocke;

	return ftp_fakereply(299, RETRY_NO,
	    "post_initconnection_active() successful");
}

typedef struct {
	char *rname;
	char *lname;
	int retrls;	//   0 = retrieve file, 1 = retrieve ls -lR
} retrieveinfo;

static ftpreply *
ftp_rawretrieve (ftpstat *fs, retrieveinfo *ri)
{
	//
	// *TODO* Currently we do only passive FTP transfers.
	//        But we cannot expect all servers to support
	//        passive FTP, so we should also support active
	//        transfers.
	//

	ftpreply *fr;
	int datasocke, retrcode, lfile;
	struct timeval xferstart, xferend;
	long numbytes;

	//
	//   First, set the transfer type to "image" (binary),
	//   or ascii, if necessary.  If we want an ls -lR,
	//   we also make sure that we are in the requested
	//   directory.
	//

	fr = ftp_needtype(fs, ri->retrls ? TYPE_ASCII : TYPE_IMAGE);
	if (fr->code >= 400)
		return fr;
	if (ri->retrls) {
		ftp_freereply (fr);
		fr = ftp_needcwd(fs, ri->rname);
		if (fr->code >= 400)
			return fr;
	}

	//
	//   Prepare the data connection.
	//

	ftp_freereply (fr);
	if ((cc->flags & MIRROR_ACTIVE))
		fr = pre_initconnection_active(fs, &datasocke);
	else
		fr = pre_initconnection_passive(fs, &datasocke);
	if (fr->code >= 400)
		return fr;

	//
	//   Initiate the file transfer.
	//

	ftp_freereply (fr);
	if (ri->retrls)
		fr = ftp_cmd(fs, "LIST -lRa");
	else
		fr = ftp_cmd(fs, "RETR %s", ri->rname);
	if (fr->code >= 400) {
		//
		//   Problem.  File not found, permission
		//   denied, or something else.
		//
		// *TODO* Some servers seem to have problems
		//        with complete path names in the RETR
		//        command.  We could try to CWD there
		//        first and do a RETR on the basename
		//        only.
		//
		close (datasocke);
		return fr;
	}
	retrcode = fr->code;

	//
	//   Finalize preparation of the data connection
	//   (only needs to be done for active FTP mode).
	//

	if ((cc->flags & MIRROR_ACTIVE)) {
		ftp_freereply (fr);
		fr = post_initconnection_active(fs, &datasocke);
		if (fr->code >= 400)
			return fr;
	}

	//
	//   Open the local file for writing, then transfer
	//   the stuff from the data connection to the file.
	//   If we're retrieving an ls -lR, we don't need a
	//   local file and transfer the result directly into
	//   memory.
	//

	gettimeofday (&xferstart, NULL);
	if (ri->retrls) {
		setsize_progress (0);
		start_progress();
		numbytes = lscopy(datasocke, (char **) ri->lname);
		stop_progress();
		if (numbytes == -1) {
			ftp_freereply (fr);
			fr = ftp_fakereply(499, RETRY_RETRY,
			    "error reading ls -lR from %s", ri->rname);
		}
		else if (numbytes == -2) {
			ftp_freereply (fr);
			fr = ftp_fakereply(599, RETRY_NO,
			    "out of memory reading ls -lR from %s", ri->rname);
		}
	}
	else if ((lfile = open(ri->lname, O_WRONLY | O_CREAT | O_TRUNC, 0600))
	    >= 0) {
		start_progress();
		numbytes = fdcopy(datasocke, lfile);
		stop_progress();
		if (numbytes == -1) {
			ftp_freereply (fr);
			fr = ftp_fakereply(499, RETRY_RETRY,
			    "error reading from %s", ri->rname);
		}
		else if (numbytes == -2) {
			ftp_freereply (fr);
			fs->locfail++;
			fr = ftp_fakereply(599, RETRY_NO,
			    "error writing to %s", ri->lname);
		}
		if (close(lfile) < 0) {
			ftp_freereply (fr);
			fr = ftp_fakereply(599, RETRY_NO,
			    "close(%s): %s", ri->lname, serror());
		}
	}
	else {
		numbytes = 0;
		ftp_freereply (fr);
		fs->locfail++;
		fr = ftp_fakereply(599, RETRY_NO,
		    "open(%s): %s", ri->lname, serror());
	}
	gettimeofday (&xferend, NULL);
	if (numbytes > 0) {
		double banana;

		TIMEVALDIFF(xferstart,xferend,xferend);
		banana = xferend.tv_sec + ((double) xferend.tv_usec / 1000000);
		if (cc->debug & DEBUG_XFERRATE) {
			fprintf (stderr, "Transfer rate:  %s\n",
			    xferrate_report(banana, numbytes));
		}
		cs->xfertime += banana;
		cs->xferbytes += numbytes;
	}
	close (datasocke);
	if (retrcode <= 199)
		if (fr->code >= 400)
			ftp_freereply(ftp_readreply(fs));
		else {
			ftp_freereply (fr);
			fr = ftp_readreply(fs);
		}
	return fr;
}

//
//   Retrieves the specified remote file to the
//   specified local file.
//

ftpreply *
ftp_retrieve (ftpstat *fs, char *rname, char *lname)
{
	retrieveinfo ri;

	ri.rname = rname;
	ri.lname = lname;
	ri.retrls = 0;
	return ftp_sequence(fs, (sequencejob *) ftp_rawretrieve, &ri);
}

//
//   Retrieves an ls -lRa from the  specified remote directory
//   and returns a pointer to it in *list.
//

ftpreply *
ftp_lslra (ftpstat *fs, char *rdir, char **list)
{
	retrieveinfo ri;

	ri.rname = rdir;
	ri.lname = (char *) list;
	ri.retrls = 1;
	return ftp_sequence(fs, (sequencejob *) ftp_rawretrieve, &ri);
}

//-- 
