//
//   ftp.h
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: ftp.h,v 1.1 1998/12/04 09:04:40 olli Exp $
//

#ifndef HAVE_FTP_H
#define HAVE_FTP_H

static const char cvsid_ftp_h[]
    = "@(#)$Id: ftp.h,v 1.1 1998/12/04 09:04:40 olli Exp $";

#include <netinet/in.h>

#define DEFAULT_FTP_PORT  21

//
//   FTP transfer types.
//

typedef char ftpxfertype;

#define TYPE_ASCII	'A'
#define TYPE_IMAGE	'I'
#define TYPE_UNKNOWN	'?'

//
//   FTP server status structure.
//

typedef struct {
	char *name;		// server name
	struct in_addr ip;	// server IP number
	unsigned short port;	// port number
	struct in_addr lip;	// local socket IP number
	int fd;			// command socket descriptor, -1 = closed
	char *cwd;		// current directory, NULL = unknown
	char *input;		// remaining server input, NULL = none
	int inputsize;		// number of bytes in *input (if != NULL)
	ftpxfertype type;	// current transfer type
	unsigned support;	// server support flags, see below
	unsigned sretry;	// number of soft retries
	unsigned hretry;	// number of hard retries
	unsigned remfail;	// number of unrecovered remote failures
	unsigned locfail;	// number of unrecovered local failures
	unsigned rodelay;	// current re-open delay
	unsigned pasvfail;	// number of PASV failures
	unsigned pasvsuccess;	// number of PASV successes
	unsigned pasvlast;	// "distance" of last successful PASV
} ftpstat;

//
//   Server support flags.  Note that not all of them are
//   supported by omi at this time.
//
//   For an "ideal" FTP server, all of these flags should be
//   set, like they are for the wuftpd.  But unfortunately, we
//   aren't living in an ideal world, and most other ftpds
//   don't support one or more of these things.  If one or more
//   of these flags are not set, we have to do certain things
//   in different ways:
//
//    - If FTPD_STAT is false, we have to use LIST instead.
//
//    - If FTPD_PASV is false, we have to use active FTP
//      transfers instead (PORT command) for LIST and RETR.
//
//    - If FTPD_ABSRETR is false, it means that we should not
//      use full paths in the RETR command (yeah, some servers
//      are that braindead).  Use CWD to go into the directory
//      and then RETR with the basename of the file only.
//
//    - If FTPD_LS_L is false, we don't use it and hope that we
//      still get a Unix-style ls -l output.
//
//    - If FTPD_LS_R is false, we CWD into every directory and
//      do a non-recursive STAT or LIST there.
//
//    - If FTPD_LS_A is false, don't use it and hope that the
//      server defaults to -a or -A (most servers do).
//
//    - If FTPD_LS_T is false, we don't use it (and in that
//      case we have to accept the reduced accuracy of time
//      stamps.)
//
//   If either of the -lRa flags is not supported, the server
//   is probably not a Unix-type server (and probably doesn't
//   accept reasonable options for STAT and LIST at all).
//   Hopefully it produces a UNIX-like ls -l output, otherwise
//   we're lost.
//   We must be careful to say "STAT ." if we don't use any
//   options, because "STAT" without any arguments doesn't get
//   us a directory listing (see RFC 959).
//

#define FTPD_STAT	0x00001	// server supports STAT
#define FTPD_PASV	0x00002	// server supports PASV
#define FTPD_ABSRETR	0x00004	// server supports absolute RETR
#define FTPD_LS_L	0x00010	// server supports ls -l
#define FTPD_LS_R	0x00020	// server supports ls -R
#define FTPD_LS_A	0x00040	// server supports ls -a
#define FTPD_LS_T	0x00080	// server supports ls -T

typedef struct {
	char *text;	// the complete reply text
	char *line;	// pointer to the text of the last line within *text
	int code;	// the FTP reply code
	int retry;	// retry status, see below
} ftpreply;

#define RETRY_NO	0	// do not retry
#define RETRY_REOPEN	1	// re-open + retry
#define RETRY_RETRY	2	// retry immediately
#define RETRY_STOP	3	// panic, pee your pants, and give up

typedef ftpreply * sequencejob (ftpstat *fs, void *data);

ftpstat  * ftp_initserver (char *hostspec, unsigned int port);
ftpreply * ftp_retrieve   (ftpstat *fs, char *rname, char *lname);
ftpreply * ftp_lslra      (ftpstat *fs, char *rdir, char **list);
ftpreply * ftp_sequence   (ftpstat *fs, sequencejob *job, void *data);
ftpreply * ftp_needtype   (ftpstat *fs, ftpxfertype reqtype);
ftpreply * ftp_needcwd    (ftpstat *fs, char *reqcwd);
ftpreply * ftp_cmd        (ftpstat *fs, char *format, /* arguments */ ...);
ftpreply * ftp_doneserver (ftpstat *fs);
ftpreply * ftp_fakereply  (int code, int retry, char *format, /* args */ ...);
void       ftp_freereply  (ftpreply *fr);

#endif // HAVE_FTP_H

//-- 
