//
//   lex.c
//
//   Oliver Frome  <olli@fromme.com>
//   @(#)$Id: lex.c,v 1.3 1998/11/02 05:03:33 olli Exp $
//
//   This module implements a tokenizer with quoting, command
//   substitution and parameter substitution.  It tries to
//   mimic the behaviour of /bin/sh as closely as possible.
//

static const char cvsid[]
    = "@(#)$Id: lex.c,v 1.3 1998/11/02 05:03:33 olli Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lex.h"

//
//   The following constants (bit fields) are used for
//   lexstate.quote_state (see below).
//
//   QUOTE_DOUBLE   is set if we're inside double quotes.
//   QUOTE_SINGLE   is set if we're inside single quotes.
//   QUOTE_BSLASH   is set if the next character is being
//                  escaped by a backslash.
//   QUOTE_BTICK    is set if we're inside backticks.
//   QUOTE_PARAM    is set if we're inside a parameter.
//   QUOTE_SUBST    is set if we're inside the result of a
//                  command substitution (backticks) or
//                  parameter substitution (dollar sign).
//   QUOTE_SPACE    is set if we encountered whitespace that
//                  is relevant for splitting/tokenizing.
//                  It is NOT set for quoted whitespace.
//

#define QUOTE_DOUBLE	0x01
#define QUOTE_SINGLE	0x02
#define QUOTE_BSLASH	0x04
#define QUOTE_BTICK	0x08
#define QUOTE_QUOTE	(QUOTE_DOUBLE | QUOTE_SINGLE \
			| QUOTE_BSLASH | QUOTE_BTICK)
#define QUOTE_PARAM	0x10
#define QUOTE_SUBST	0x20
#define QUOTE_SPACE	0x40

int lineno;		// current line number
lex_inputfunc  lif;	// lex() input function, set with lex_setinput()
lex_outputfunc lof;	// lex() output function, set with lex_setoutput()
void *lifudata;		// user data for input function
void *lofudata;		// user data for output function

void
lex_setinput (lex_inputfunc func, void *userdata)
{
	lif = func;
	lifudata = userdata;
}

void
lex_setoutput (lex_outputfunc func, void *userdata)
{
	lof = func;
	lofudata = userdata;
}

static int
isspace (char c)
{
	return c == ' ' || c == '\t' || c == '\n';
}

static int
isparamstart (char c)
{
	if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || c == '_')
		return 1;
	return 0;
}

static int
isparamchar (char c)
{
	if (isparamstart(c) || (c >= '0' && c <= '9'))
		return 1;
	return 0;
}

//
//   strappend(s, a) works similar to strcat(s, a), but treats
//   "s" as a dynamically allocated string.
//
//   The return value is the new location of the resulting
//   string (or NULL if an error occured), the value of "s" is
//   undefined in either case after calling strappend().
//
//   chrappend(s, a) works similar but appends only a single
//   character.
//

static char *
strappend (char *s, char *a)
{
	char *new;
	int sl, al;

	sl = s ? strlen(s) : 0;
	al = strlen(a);
	if (!(new = (char *) realloc(s, sl + al + 1))) {
		free (s);
		return NULL;
	}
	if (s)
		strcpy (new, s);
	strcpy (new + sl, a);
	return new;
}

static char *
chrappend (char *s, char a)
{
	char *new;
	int sl;

	sl = s ? strlen(s) : 0;
	if (!(new = (char *) realloc(s, sl + 2))) {
		free (s);
		return NULL;
	}
	if (s)
		strcpy (new, s);
	new[sl] = a;
	new[sl + 1] = '\0';
	return new;
}

static void
freetokens (int numtokens, char *tokens[])
{
	if (numtokens && tokens) {
		for (; numtokens; numtokens--)
			free (tokens[numtokens - 1]);
		free (tokens);
	}
}

static int
addtoken (int *numtokens, char **tokens[], char **token)
{
	char **newtoks;

	if (!(newtoks = (char **) realloc(*tokens,
	    sizeof(char *) * (*numtokens + 1)))) {
		freetokens (*numtokens, *tokens);
		free (*token);
		*token = NULL;
		return 0;
	}
	if (!(newtoks[*numtokens] = *token ? strdup(*token) : strdup(""))) {
		freetokens (*numtokens + 1, newtoks);
		free (*token);
		*token = NULL;
		return 0;
	}
	*numtokens += 1;
	free (*token);
	*token = NULL;
	*tokens = newtoks;
	return 1;
}

char *
readinput (FILE *inputfile)
{
	char *line, *last;
	char buf[256];

	line = NULL;
	for (;;) {
		if (!fgets(buf, 256, inputfile))
			break;
		if (!(line = strappend(line, buf)))
			return NULL;
		last = line + (strlen(line) - 1);
		if (*last == '\n') {
			*last = '\0';
			break;
		}
	}
	lineno++;
	return line;
}

typedef struct lexstate_t {
	char *token;		// the current token
	char **tokens;		// list of tokens
	int numtokens;		// number of tokens in the list
	int quote_state;	// current quoting state
	char *cmdpar;		// parameter or backtick contents
} lexstate;

static int lexcommand (lexstate *ls, char *src);

static int lexparameter (lexstate *ls, char *src);

static int
lexchar (lexstate *ls, char *src)
{
	//   Inside single quotes, everything is quoted
	//   except for the terminating single quote.
	if ((ls->quote_state & QUOTE_SINGLE))
		if (*src == '\'') {
			ls->quote_state &= ~QUOTE_SINGLE;
			return 1;
		}
		else
			goto copychar;

	//   Inside backticks (``), just collect characters
	//   into the command string, until we find an unquoted
	//   backtick that terminates the command string.
	if ((ls->quote_state & QUOTE_BTICK))
		return lexcommand(ls, src);

	//   Inside backticks (``), just collect characters
	//   into the command string, until we find an unquoted
	//   backtick that terminates the command string.
	if ((ls->quote_state & QUOTE_PARAM))
		return lexparameter(ls, src);

	//   If the previous character was an unquoted backslash,
	//   this character is quoted, no matter what it is.
	//   (Well, if it's a newline, we remove it.)
	if ((ls->quote_state & QUOTE_BSLASH)) {
		ls->quote_state &= ~QUOTE_BSLASH;
		if (*src == '\n')
			return 1;
		goto copychar;
	}

	//   Input is split into tokens at sequences of
	//   unquoted whitespace.
	if (isspace(*src)) {
		if ((ls->quote_state & QUOTE_DOUBLE))
			goto copychar;
		else if ((ls->quote_state & QUOTE_SPACE))
			return 1;
		if (!(addtoken(&ls->numtokens, &ls->tokens, &ls->token)))
			return 0;
		ls->quote_state |= QUOTE_SPACE;
		return 1;
	}

	//   At this point, if we're inside the result of a
	//   command substitution (resulted from backticks),
	//   everything is taken verbatim.
	//   Note that whitespace-splitting does take place,
	//   unless the whole thing is also in double quotes
	//   (see above).
	if ((ls->quote_state & QUOTE_SUBST)) {
		ls->quote_state &= ~QUOTE_SPACE;
		goto copychar;
	}

	//   A word beginning with an unquoted '#' introduces
	//   a comment that extends to the end of the line.
	if (*src == '#' && (ls->quote_state & QUOTE_SPACE))
		return 2;

	//   If a backslash (unquoted or within double quotes) is
	//   followed by a newline, both characters are removed.
	//   An unquoted backslash quotes the following character,
	//   no matter what it is.  However, within double quotes,
	//   a backslash quotes the following character only if it's
	//   "`", '"', '$' or another backslash; otherwise the
	//   backslash is taken literally.
	if (*src == '\\') {
		//   If a backslash (unquoted or within double quotes)
		//   is followed by a newline, both characters are
		//   removed.  Note that the QUOTE_SPACE state (set
		//   or not set) may not change at this point.
		if (!src[1] || src[1] == '\n') {
			ls->quote_state |= QUOTE_BSLASH;
			return 1;
		}
		if ((ls->quote_state & QUOTE_DOUBLE) && src[1] != '\\'
		    && src[1] != '`' && src[1] != '\"' && src[1] != '$')
			goto copychar;
		ls->quote_state &= ~QUOTE_SPACE;
		ls->quote_state |= QUOTE_BSLASH;
		return 1;
	}

	//   Starting backtick for command substitution.  (The
	//   terminating backtick is handled at the beginning.)
	if (*src == '`') {
		ls->quote_state |= QUOTE_BTICK;
		if (ls->cmdpar)
			free (ls->cmdpar);
		ls->cmdpar = NULL;
		return 1;
	}

	//   A dollar sign introduces a parameter substitution,
	//   unless followed by an invalid character.
	if (*src == '$' && (isparamstart(src[1]) || src[1] == '{')) {
		ls->quote_state |= QUOTE_PARAM;
		if (ls->cmdpar)
			free (ls->cmdpar);
		ls->cmdpar = NULL;
		return 1;
	}

	//   At this point, we definitely have something which is not
	//   unquoted whitespace, thus reset the QUOTE_SPACE flag.
	ls->quote_state &= ~QUOTE_SPACE;

	//   Starting or terminating double quotes.
	if (*src == '\"') {
		ls->quote_state ^= QUOTE_DOUBLE;
		return 1;
	}

	//   Starting single quote.  (Terminating single quote
	//   is handled at the beginning.)
	if (*src == '\'') {
		if ((ls->quote_state & QUOTE_DOUBLE))
			goto copychar;
		ls->quote_state |= QUOTE_SINGLE;
		return 1;
	}

	//   Append the current character literally to
	//   the current token.
	copychar:
	if (!(ls->token = chrappend(ls->token, *src)))
		return 0;
	return 1;
}

static int
lexcommand (lexstate *ls, char *src)
{
	if ((ls->quote_state & QUOTE_BSLASH))
		ls->quote_state &= ~QUOTE_BSLASH;
	else if (*src == '\\') {
		if (!src[1] || src[1] == '$'
		    || src[1] == '`' || src[1] == '\\') {
			ls->quote_state |= QUOTE_BSLASH;
			return 1;
		}
		if ((ls->quote_state & QUOTE_DOUBLE) && src[1] == '\"')
			return 1;
	}
	else if (*src == '`') {
		FILE *pipe;
		char *pline, *cptr;
		int oldlineno, result;

		ls->quote_state &= ~QUOTE_BTICK;
		if (!(pipe = popen(ls->cmdpar, "r"))) {
			fprintf (stderr, "Error in line %d:  "
			    "Can't execute command.\n", lineno);
			return 0;
		}
		oldlineno = lineno;
		ls->quote_state |= QUOTE_SUBST;
		result = 1;
		while (result && (pline = readinput(pipe))) {
			if (result == 2) {
				if (!lexchar(ls, "\n")) {
					result = 0;
					free (pline);
					break;
				}
			}
			else
				result = 2;
			for (cptr = pline; *cptr; cptr++)
				if (!lexchar(ls, cptr)) {
					result = 0;
					break;
				}
			free (pline);
		}
		ls->quote_state &= ~QUOTE_SUBST;
		lineno = oldlineno;
		pclose (pipe);
		free (ls->cmdpar);
		ls->cmdpar = NULL;
		return result ? 1 : 0;
	}
	if (!(ls->cmdpar = chrappend(ls->cmdpar, *src)))
		return 0;
	return 1;
}

static int
lexparameter (lexstate *ls, char *src)
{
	if (*src == '}') {
		//
		//   If this is the closing brace, we're
		//   already done.  Just clear the flag
		//   and return.
		//
		ls->quote_state &= ~QUOTE_PARAM;
		return 1;
	}
	if (!(ls->cmdpar = chrappend(ls->cmdpar, *src)))
		return 0;
	//
	//   If the next character does not belong to the
	//   parameter anymore, then perform the actual
	//   parameter substitution.
	//
	if (!isparamchar(src[1])) {
		char *value;
		int result;

		ls->quote_state &= ~QUOTE_PARAM;
		result = 1;
		if (*(value = ls->cmdpar) == '{')
			value++;
		if ((value = getenv(value))) {
			ls->quote_state |= QUOTE_SUBST;
			for (; *value; value++)
				if (!lexchar(ls, value)) {
					result = 0;
					break;
				}
			ls->quote_state &= ~QUOTE_SUBST;
		}
		//
		//   If the parameter is enclosed in braces,
		//   set the QUOTE_PARAM flag again, so we
		//   will swallow the '}' at the next call.
		//
		if (src[1] == '}' && *ls->cmdpar == '{')
			ls->quote_state |= QUOTE_PARAM;
		free (ls->cmdpar);
		ls->cmdpar = NULL;
		return result ? 1 : 0;
	}
	return 1;
}

static int
lex_quotenewline (lexstate *ls, char **line)
{
	//   nl     (unquoted newline terminates the logical line)
	//   "nl    -->   nl
	//   'nl    -->   nl
	//   `nl    -->   nl
	//   \nl    -->   (nothing)
	//   "\nl   -->   (nothing)
	//   '\nl   (handled by the 'nl case, the \ is literal)
	//   `\nl   -->   (nothing)

	if ((ls->quote_state & QUOTE_BTICK)
	    && !(ls->quote_state & QUOTE_BSLASH))
	{
		if (!(ls->cmdpar = chrappend(ls->cmdpar, '\n')))
			return 0;
	}
	else if ((ls->quote_state & QUOTE_SINGLE) ||
	    ((ls->quote_state & QUOTE_DOUBLE)
	    && !(ls->quote_state & QUOTE_BSLASH)))
	{
		if (!(ls->token = chrappend(ls->token, '\n')))
			return 0;
	}
	ls->quote_state &= ~QUOTE_BSLASH;
	*line = lif(lifudata);	//   fetch another physical line
	return 1;
}

static int
lex_quotecheck (lexstate *ls)
{
	if ((ls->quote_state & QUOTE_DOUBLE)) {
		fprintf (stderr,
		    "Error in line %d:  Open double quote.\n", lineno);
		return 0;
	}
	if ((ls->quote_state & QUOTE_SINGLE)) {
		fprintf (stderr,
		    "Error in line %d:  Open single quote.\n", lineno);
		return 0;
	}
	if ((ls->quote_state & QUOTE_BSLASH)) {
		fprintf (stderr,
		    "Error in line %d:  Open backslash quote.\n", lineno);
		return 0;
	}
	if ((ls->quote_state & QUOTE_BTICK)) {
		fprintf (stderr,
		    "Error in line %d:  Open command substitution.\n",
		    lineno);
		return 0;
	}
	return 1;
}

int
lex (void)
{
	char *line, *src;
	lexstate ls;
	int result = 1;

	while ((line = lif(lifudata))) {
		ls.token = NULL;
		ls.tokens = NULL;
		ls.numtokens = 0;
		ls.quote_state = QUOTE_SPACE;
		ls.cmdpar = NULL;
		while (1) {
			for (src = line; *src; src++) {
				if ((result = lexchar(&ls, src)) < 1)
					return 0;
				if (result == 2)	// comment
					break;
			}
			free (line);
			if ((ls.quote_state & QUOTE_QUOTE)) {
				if (!lex_quotenewline(&ls, &line))
					return 0;
				if (line)
					continue;
			}
			break;
		}
		if (ls.token || !(ls.quote_state & QUOTE_SPACE))
			if (!(addtoken(&ls.numtokens, &ls.tokens, &ls.token)))
				return 0;
		result = lex_quotecheck(&ls);
		if (!lof(ls.numtokens, ls.tokens, lofudata)) {
			freetokens (ls.numtokens, ls.tokens);
			return 0;
		}
		freetokens (ls.numtokens, ls.tokens);
	}
	return result;
}

#ifdef DEBUG_STANDALONE

int
writeoutput (int tokenc, char *tokenv[], void *dummy)
{
	int i;

	for (i = 0; i < tokenc; i++)
		printf ("#%d: \"%s\"\n", i, tokenv[i]);
	printf ("\n");
	return 1;
}

int
main (int argc, char *argv[])
{
	lineno = 0;
	lex_setinput ((lex_inputfunc) readinput, stdin);
	lex_setoutput ((lex_outputfunc) writeoutput, NULL);
	return lex() == 0;
}

#endif

/*

Excerpt from Solaris 2.6 "man sh"
=================================

  Comments Lines
     A word beginning with # causes that word and all the follow-
     ing characters up to a newline to be ignored.

  Command Substitution
     The shell reads commands from the string between  two  grave
     accents  (` `)  and  the standard output from these commands
     may be used as all or part of  a  word.   Trailing  newlines
     from the standard output are removed.

     No interpretation is done on the string before the string is
     read,  except to remove backslashes (\) used to escape other
     characters.  Backslashes may  be  used  to  escape  a  grave
     accent  (`)  or another backslash (\) and are removed before
     the command string is read.  Escaping grave  accents  allows
     nested  command  substitution.   If the command substitution
     lies within a pair of double quotes (" ... `...` ...  "),  a
     backslash  used  to  escape  a  double  quote  (\")  will be
     removed; otherwise, it will be left intact.

     If a backslash is used to escape a newline character  (\new-
     line),  both  the backslash and the newline are removed (see
     the later section on  Quoting).   In  addition,  backslashes
     used  to  escape  dollar  signs  (\$) are removed.  Since no
     parameter substitution is done on the command string  before
     it  is  read,  inserting a backslash to escape a dollar sign
     has no effect.  Backslashes that  precede  characters  other
     than  \, `, ", newline,  and $ are left intact when the com-
     mand string is read.

  Parameter Substitution
     The  character  $  is  used   to   introduce   substitutable
     parameters.   [...]

     ${parameter}
          The value, if any, of  the  parameter  is  substituted.
          The braces are required only when parameter is followed
          by a letter, digit, or underscore that  is  not  to  be
          interpreted  as part of its name.  [...]

     The following parameters are automatically set by the shell.
          [...]
          $    The process number of this shell.

     [...]

  Blank Interpretation
     After parameter and command  substitution,  the  results  of
     substitution  are scanned for internal field separator char-
     acters (those found in IFS) and split  into  distinct  argu-
     ments  where such characters are found.  Explicit null argu-
     ments ("" or '')  are  retained.   Implicit  null  arguments
     (those  resulting  from  parameters that have no values) are
     removed.

  Quoting
     The following characters have a special meaning to the shell
     and cause termination of a word unless quoted:

          ;  &  (  )  |  ^  <  >  newline  space  tab

     A character may be  quoted  (that  is,  made  to  stand  for
     itself) by preceding it with a backslash (\) or inserting it
     between a pair of quote marks ('' or "").   During  process-
     ing,  the shell may quote certain characters to prevent them
     from taking on a special meaning.  Backslashes used to quote
     a single character are removed from the word before the com-
     mand is executed.  The pair \newline is removed from a  word
     before command and parameter substitution.

     All characters enclosed between a pair of single quote marks
     (''),  except  a  single  quote,  are  quoted  by the shell.
     Backslash has no special meaning inside  a  pair  of  single
     quotes.   A single quote may be quoted inside a pair of dou-
     ble quote marks (for example, "'"), but a single  quote  can
     not be quoted inside a pair of single quotes.

     Inside a pair of double quote marks (""), parameter and com-
     mand substitution occurs and the shell quotes the results to
     avoid blank interpretation and file name generation.  If  $*
     is within a pair of double quotes, the positional parameters
     are substituted and quoted, separated by quoted spaces  ("$1
     $2  ..."); however, if $@ is within a pair of double quotes,
     the  positional  parameters  are  substituted  and   quoted,
     separated by unquoted spaces ("$1" "$2" ... ).  \ quotes the
     characters \, `, ", and $.  The  pair  \newline  is  removed
     before  parameter  and command substitution.  If a backslash
     precedes characters other than \, `, ", $, and newline, then
     the backslash itself is quoted by the shell.

*/

//-- 
