//
//   lex.h
//
//   Oliver Frome  <olli@fromme.com>
//   @(#)$Id: lex.h,v 1.2 1998/12/04 09:04:40 olli Exp $
//
//   This module implements a tokenizer with quoting, command
//   substitution and parameter substitution.  It tries to
//   mimic the behaviour of /bin/sh as closely as possible.
//

#ifndef HAVE_LEX_H
#define HAVE_LEX_H

static const char cvsid_lex_h[]
    = "@(#)$Id: lex.h,v 1.2 1998/12/04 09:04:40 olli Exp $";

//
//   The input function to lex() is expected to return
//   exactly one physical line (not including the newline
//   character) from the input, or NULL upon EOF or if an
//   error occured.  It should also update the global
//   lineno variable.  The calling function is responsible
//   for freeing the line when it's not needed anymore.
//

typedef char * (*lex_inputfunc) (void *userdata);

//
//   The output function for lex() is called whenever a
//   logical line has been parsed and tokenized successfully.
//

typedef int (*lex_outputfunc) (int tokenc, char *tokenv[], void *userdata);

extern int lineno;	// current line number

void lex_setinput (lex_inputfunc func, void *userdata);
void lex_setoutput (lex_outputfunc func, void *userdata);
char * readinput (FILE *inputfile);

//
//   The lex() function works like this:
//
//   1)  Read and parse an entire line, i.e. until an unquoted
//       (literal) newline is encountered, or EOF.  Note that
//       such a line can consist of multiple physical lines if
//       they're separated by quoted newlines.
//
//       While parsing the line, quoting and spacing is handled
//       appropriately.  The quoting is intended to be similar
//       to the bourne shell.  The line is split into tokens at
//       unquoted whitespace.  If there's an open quote when
//       EOF is encountered, an error is returned.
//
//   2)  The resulting list of tokens (of which the first one
//       is the "key", and the remaining ones are the "value")
//       is then handed to the output function.
//
//   3)  Continue at step 1, until EOF is encountered.
//
//   First, we define a lexchar() function which handles one
//   character at a time (but needs to be able to look at the
//   following characters of the line, too).
//
//   The function lex_quotenewline() handles the case of an
//   open quote (single, double, or backslash) at the end of a
//   physical line.
//

int lex (void);

#endif // HAVE_LEX_H

//-- 
