//
//   lst.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: lst.c,v 1.6 1998/12/04 09:04:42 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: lst.c,v 1.6 1998/12/04 09:04:42 olli Exp $";

#include <string.h>
#include <stdlib.h>

#include "lst.h"
#include "utils.h"

//
//   The following functions work on simple, unordered lists.
//

void
simple_list_init (simple_list *l)
{
	l->count = 0;
	l->list = NULL;
}

void
simple_list_add (simple_list *l, void *new, int size)
{
	if (!(l->list = realloc(l->list, (l->count + 1) * sizeof(*l->list))))
		out_of_memory();
	if (new && size) {
		l->list[l->count] = tmalloc(size);
		memcpy (l->list[l->count++], new, size);
	}
	else
		l->list[l->count++] = NULL;
}

int
simple_list_count (simple_list *l)
{
	return l->count;
}

//
//   Traverse a simple list.
//
//   If the job function returns a non-zero value, the
//   traversal is aborted, and the function returns with
//   that value.  If all job functions return zero, the
//   return value is also zero.
//
//   If TRAVERSE_IGNORERESULT is set in the flags, the
//   return values of the job functions are ignored, and
//   the return value is always zero.
//

int
simple_list_traverse (simple_list *l, jobfunc *job, void *jobdata,
    unsigned flags)
{
	int i, result;

	if (!l)
		return 0;
	if (flags & TRAVERSE_IGNORERESULT) {
		for (i = 0; i < l->count; i++)
			if (l->list[i])
				job (l->list[i], jobdata);
	}
	else
		for (i = 0; i < l->count; i++)
			if (l->list[i])
				if ((result = job(l->list[i], jobdata)))
					return result;
	return 0;
}

void
simple_list_done (simple_list *l)
{
	int i;

	for (i = 0; i < l->count; i++)
		if (l->list[i])
			free (l->list[i]);
	free (l->list);
	l->count = 0;
	l->list = NULL;
}

//
//   The following functions work on CSV strings
//   (CSV = comma-separated values).
//
//   csv_count() returns the number of values in a CSV string.
//               Note that values can also be empty, i.e.
//               csv_count("x,,y,") == 4 && csv_count("") == 1.
//
//   csv_value() returns a zero-terminated copy of the n'th
//               value in a CSV string, starting with n = 0.
//               Returns NULL if n >= csv_count().
//               Note that it's the responsibility of the
//               calling function to free() the copy of the
//               value.
//
//   csv_traverse() calls the given job function for each value
//                  of the given CSV string.
//
//   NB:  We're not using strtok() because of several
//   disadvantages, and we're not using strsep() either
//   because it's not portable enough.
//

int
csv_count (char *csv)
{
	int i;

	i = 1;
	while (*csv)
		if (*csv++ == ',')
			i++;
	return i;
}

char *
csv_value (char *csv, int n)
{
	char *v;

	if (n) {
		while (*csv)
			if (*csv++ == ',')
				if (!--n)
					break;
		if (n)
			return NULL;
	}
	if (!(v = strndup(csv, strcspn(csv, ","))))
		out_of_memory();
	return justify(v);
}

int
csv_traverse (char *csv, jobfunc *job, void *jobdata, unsigned flags)
{
	char *v;
	int len, result;

	for (;;) {
		len = strcspn(csv, ",");
		if (!(v = strndup(csv, len)))
			out_of_memory();
		result = job(justify(v), jobdata);
		free (v);
		if (!result && !(flags & TRAVERSE_IGNORERESULT))
			return FALSE;
		if (!*(csv += len))
			break;
		csv++;
	}
	return TRUE;
}

//-- 
