//
//   pro.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: pro.c,v 1.6 1998/12/12 22:23:16 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: pro.c,v 1.6 1998/12/12 22:23:16 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

extern int errno;

#include "utils.h"
#include "pro.h"
#include "stt.h"

//
//   Activate the progress display if a transfer takes longer
//   than 10 seconds, and update the display every 5 seconds.
//   Declare the transfer as "stalled" if it has been dead for
//   at least 10 seconds.
//

#define PROGRESS_ACTIVATE  10
#define PROGRESS_UPDATE    5
#define PROGRESS_STALLED   10

int procmd = -1;
pid_t propid = -1;

static int progress_main (FILE *outputfile);

void
init_progress (void)
{
	int propipe[2];
	FILE *output;

	if (isatty(fileno(stdout)))
		output = stdout;
	else if (isatty(fileno(stderr)))
		output = stderr;
	else {
		procmd = propid = -1;
		return;
	}

	//
	//   Create a pipe which is used to send commands to
	//   the child process to handle the progress display.
	//   Then fork the child process.
	//

	if (pipe(propipe) < 0) {
		fprintf (stderr, "%s: pipe(): %s\n", me, serror());
		fprintf (stderr, "pipe() failed, progress display disabled.\n");
		procmd = propid = -1;
		return;
	}
	fflush (output);
	switch ((propid = fork())) {
	case -1:
		//
		//   error
		//
		close (propipe[0]);
		close (propipe[1]);
		fprintf (stderr, "%s: fork(): %s\n", me, serror());
		fprintf (stderr, "fork() failed, progress display disabled.\n");
		procmd = propid = -1;
		return;
	case 0:
		//
		//   child process
		//
		if (propipe[0] != STDIN_FILENO) {
			if (dup2(propipe[0], STDIN_FILENO) < 0) {
				//   Uh oh...
				fprintf (stderr, "%s: dup2(): %s\n",
				    me, serror());
				_exit (127);
			}
			close (propipe[0]);
		}
		close (propipe[1]);
		_exit (progress_main(output));
	}
	//
	//   parent
	//
	procmd = propipe[1];
	close (propipe[0]);
}

int
done_progress (void)
{
	sigset_t newsigmask, oldsigmask;
	int stat;
	pid_t pid;

	if (procmd >= 0) {
		close (procmd);
		procmd = -1;
	}
	if (propid > 0) {
		sigemptyset (&newsigmask);
		sigaddset (&newsigmask, SIGINT);
		sigaddset (&newsigmask, SIGQUIT);
		sigaddset (&newsigmask, SIGHUP);
		sigprocmask (SIG_BLOCK, &newsigmask, &oldsigmask);
		do {
			pid = waitpid(propid, &stat, 0);
		} while (pid == -1 && errno == EINTR);
		sigprocmask (SIG_SETMASK, &oldsigmask, NULL);
		propid = -1;
		return pid == -1 ? -1 : stat;
	}
	return 0;
}

void
command_progress (ulonglong tok, ulonglong par)
{
	ulonglong cmd[2];

	if (procmd >= 0) {
		cmd[0] = tok;
		cmd[1] = par;
		write (procmd, cmd, sizeof(cmd));
	}
}

//
// ============================================================
//
//  The following functions are executed by the child process.
//
// ============================================================
//

struct timeval prostart, prochange;
ulonglong fsize, fhave, lasthave;
FILE *output;

static void
progress_update (void)
{
	struct timeval pronow;
	int stalled, eta;
	double elapsed;

	gettimeofday (&pronow, NULL);
	if (fhave != lasthave) {
		lasthave = fhave;
		prochange = pronow;
		stalled = 0;
	}
	else
		stalled = pronow.tv_sec - prochange.tv_sec;
	TIMEVALDIFF(prostart,pronow,pronow);
	elapsed = pronow.tv_sec + ((double) pronow.tv_usec / 1000000);

	//
	//   The typecasts here are _ugly_.  Unfortunately,
	//   there is no portable way to specify formats
	//   for "long long" or something like that.
	//   This needs to be fixed somehow.
	//
	if (fsize)
		fprintf (output, " %8lu (%2d%%)",
		    (unsigned long) fhave, (int) ((fhave * 100) / fsize));
	else
		fprintf (output, " %8lu", (unsigned long) fhave);
	if (stalled >= PROGRESS_STALLED)
		fprintf (output, ", stalled for %s               \r",
		    time_format(stalled));
	else {
		if (fsize && fhave > 1) {
			eta = (elapsed * fsize) / fhave - elapsed;
			if (eta < 0)
				eta = 0;
			fprintf (output, ", %s,  %s ETA     \r",
			    xferrate_report(elapsed, fhave),
			    time_format(eta));
		}
		else
			fprintf (output, ", %s     \r",
			    xferrate_report(elapsed, fhave));
	}
	fflush (output);
}

static void
progress_clean (void)
{
	fprintf (output, "%60s\r", "");
	fflush (output);
}

static volatile sig_atomic_t have_alarm;

static void
catch_alarm (int signum)
{
	have_alarm = TRUE;
}

static int
progress_main (FILE *outputfile)
{
	fd_set rfdset;
	int input, result, interval;
	bool hadoutput;
	struct timeval timeout;
	ulonglong cmd[2];

	output = outputfile;
	input = fileno(stdin);
	alarm (0);
	catchsignal (SIGALRM, catch_alarm);
	interval = 0;
	hadoutput = FALSE;
	FD_ZERO (&rfdset);
	for (;;) {
		FD_SET (input, &rfdset);
		if (interval) {
			timeout.tv_sec = interval;
			timeout.tv_usec = 0;
			result = select(input + 1, &rfdset, NULL, NULL,
			    &timeout);
			if (result == 0)
				have_alarm = TRUE;
		}
		else
			result = select(input + 1, &rfdset, NULL, NULL, NULL);
		if (have_alarm && interval) {
			progress_update();
			hadoutput = TRUE;
			alarm (interval = PROGRESS_UPDATE);
			have_alarm = FALSE;
		}
		else if (result > 0 && FD_ISSET(input, &rfdset)) {
			if (!read(input, cmd, sizeof(cmd))) {
				alarm (0);
				if (hadoutput) {
					progress_clean ();
					hadoutput = FALSE;
				}
				break;
			}
			switch (cmd[0]) {
			case PROCMD_SETSIZE:
				fsize = cmd[1];
				break;
			case PROCMD_START:
				alarm (interval = PROGRESS_ACTIVATE);
				have_alarm = FALSE;
				fhave = 0;
				lasthave = 1;
				gettimeofday (&prostart, NULL);
				prochange = prostart;
				break;
			case PROCMD_UPDATE:
				fhave = cmd[1];
				break;
			case PROCMD_STOP:
				alarm (0);
				if (hadoutput) {
					progress_clean();
					hadoutput = FALSE;
				}
				interval = 0;
				break;
			default:	//   ?!?
				fprintf (stderr, "%s: progress: invalid "
				    "command %lu\n", me,
				    (unsigned long) cmd[0]);
				return 1;
			}
		}
		else if (result < 0) {
			//   ?!?
			fprintf (stderr, "%s: progress: select(): %s\n", me,
			    result < 0 ? serror() : "failure");
			sleep (1);
		}
	}
	return 0;
}

//-- 
