//
//   rex.c
//
//   Oliver Fromme  <olli@fromme.com>
//   @(#)$Id: rex.c,v 1.7 1998/12/12 22:28:56 olli Exp $
//

static const char cvsid[]
    = "@(#)$Id: rex.c,v 1.7 1998/12/12 22:28:56 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <sys/wait.h>

#include "utils.h"
#include "rex.h"

static void
regex_error (const int ec, const regex_t *cre)
{
	char *errmsg;
	int errsize;

	if (!(errsize = regerror(ec, cre, NULL, 0))) {
		fprintf (stderr,
		    "%s: regerror() failed!  Error code was %d.\n", me, ec);
		return;
	}
	errmsg = tmalloc(errsize);
	if (regerror(ec, cre, errmsg, errsize) > errsize)
		fprintf (stderr,
		    "%s: regerror() lost part of error message!\n", me);
	fprintf (stderr, "%s\n", errmsg);
	free (errmsg);
}

bool
add_regex (simple_list *l, const char *re, unsigned int flags)
{
	regex_t cre;	//   the compiled RE
	bregex bre;
	int result;

	bre.flags = flags;
	if (re[0] == '!') {
		bre.flags |= RE_NEGATE;
		re++;
	}
	else if (re[0] == '\\' && re[1] == '!')
		re++;
	if ((result = regcomp(&cre, re, REG_EXTENDED | REG_NOSUB))) {
		regex_error (result, &cre);
		return FALSE;
	}
	bre.regex = tmalloc(sizeof(cre));
	memcpy (bre.regex, &cre, sizeof(cre));
	simple_list_add (l, &bre, sizeof(bre));
	return TRUE;
}

bool
regex_job (bregex *bre, const char *string)
{
	int result, match;

	switch ((result = regexec(bre->regex, string, 0, NULL, 0))) {
	case 0:	//   match
		match = !(bre->flags & RE_NEGATE);
		break;
	case REG_NOMATCH:	//   mismatch
		match = bre->flags & RE_NEGATE;
		break;
	default:
		regex_error (result, bre->regex);
		fprintf (stderr,
		    "This shouldn't happen.  Time to eject a core...\n");
		abort();
	}
	if (match)
		if (bre->flags & RE_INCLUDE)
			return MATCH_INCLUDE;
		else
			return MATCH_EXCLUDE;
	else
		return MATCH_NONE;
}

// ============================================================

//
//   The following code isn't used yet.
//

#if 0

static void
pppclean (int *fdin, int *fdout, int *fderr)
{
	if (fdin[0] >= 0)	close (fdin[0]);
	if (fdin[1] >= 0)	close (fdin[1]);
	if (fdout[0] >= 0)	close (fdout[0]);
	if (fdout[1] >= 0)	close (fdout[1]);
	if (fderr[0] >= 0)	close (fderr[0]);
	if (fderr[1] >= 0)	close (fderr[1]);
}

#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif
#ifndef STDERR_FILENO
#define STDERR_FILENO 2
#endif

//
//   pppopen() is similar to popen(3), but returns the
//   filedescriptors (FDs) of stdin, stdout, and stderr
//   of the child process, and returns its process ID.
//   Returns -1 on error.
//

pid_t
pppopen(const char *command, int *fdin, int *fdout, int *fderr)
{
	int pfdin[2], pfdout[2], pfderr[2];
	pid_t child;

	pfdin[0] = pfdout[0] = pfderr[0] = -1;
	pfdin[1] = pfdout[1] = pfderr[1] = -1;
	if (fdin)
		if (pipe(pfdin) < 0)
			return -1;
	if (fdout)
		if (pipe(pfdout) < 0) {
			pppclean (pfdin, pfdout, pfderr);
			return -1;
		}
	if (fderr)
		if (pipe(pfderr) < 0) {
			pppclean (pfdin, pfdout, pfderr);
			return -1;
		}
	switch ((child = vfork())) {
	case -1:
		//
		//   error
		//
		pppclean (pfdin, pfdout, pfderr);
		return -1;
	case 0:
		//
		//   child process
		//
		if (fdin) {
			if (pfdin[0] != STDIN_FILENO) {
				if (dup2(pfdin[0], STDIN_FILENO) < 0)
					_exit (127);
				close (pfdin[0]);
			}
			close (pfdin[1]);
		}
		if (fdout) {
			if (pfdout[1] != STDOUT_FILENO) {
				if (dup2(pfdout[1], STDOUT_FILENO) < 0)
					_exit (127);
				close (pfdout[1]);
			}
			close (pfdout[0]);
		}
		if (fderr) {
			if (pfderr[1] != STDERR_FILENO) {
				if (dup2(pfderr[1], STDERR_FILENO) < 0)
					_exit (127);
				close (pfderr[1]);
			}
			close (pfderr[0]);
		}
		execl ("/bin/sh", "sh", "-c", command, NULL);
		_exit (127);
	}
	//
	//   parent
	//
	if (fdin) {
		*fdin = pfdin[1];
		close (pfdin[0]);
	}
	if (fdout) {
		*fdout = pfdout[0];
		close (pfdout[1]);
	}
	if (fderr) {
		*fderr = pfderr[0];
		close (pfderr[1]);
	}
	return child;
}

//
//   pppclose() close the given file descriptors, waits for
//   the child process to exit and returns its exit status
//   (-1 if some error occured).
//

int
pppclose (pid_t child, int fdin, int fdout, int fderr)
{
	sigset_t newsigmask, oldsigmask;
	int stat;
	pid_t pid;

	if (fdin >= 0)	close (fdin);
	if (fdout >= 0)	close (fdout);
	if (fderr >= 0)	close (fderr);
	sigemptyset (&newsigmask);
	sigaddset (&newsigmask, SIGINT);
	sigaddset (&newsigmask, SIGQUIT);
	sigaddset (&newsigmask, SIGHUP);
	sigprocmask (SIG_BLOCK, &newsigmask, &oldsigmask);
	do {
		pid = waitpid(child, &stat, 0);
	} while (pid == -1 && errno == EINTR);
	sigprocmask (SIG_SETMASK, &oldsigmask, NULL);
	return (pid == -1 ? -1 : stat);
}

#endif

//-- 
